from __future__ import annotations
import typing

from vuakhter.base.requests_log import RequestsLog
from vuakhter.kibana.elastic_log import ElasticLog
from vuakhter.utils.kibana import gen_request_entries

if typing.TYPE_CHECKING:
    from vuakhter.utils.types import RequestEntry


class ElasticRequestsLog(ElasticLog, RequestsLog):
    def __init__(self, index_pattern: str = 'django-*', *args: typing.Any, **kwargs: typing.Any):
        super().__init__(index_pattern, *args, **kwargs)

    def gen_entries(self, index: str, **kwargs: typing.Any) -> typing.Iterator[RequestEntry]:
        request_ids = kwargs.pop('request_ids', None)
        if request_ids:
            yield from gen_request_entries(self.client, index, request_ids)



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from vuakhter.utils.types import RequestEntry


class RequestsLog:
    def get_records(self, start_ts: int, end_ts: int, **kwargs: typing.Any) -> typing.Iterator[RequestEntry]:
        raise NotImplementedError()



from __future__ import annotations
import typing

from vuakhter.utils.helpers import timestamp

if typing.TYPE_CHECKING:
    from vuakhter.base.access_log import AccessLog
    from vuakhter.metrics.base import StatisticsMetrics
    from vuakhter.utils.types import DateOrDatetime

    MetricsIterable = typing.Iterable[StatisticsMetrics]
    MetricsList = typing.List[StatisticsMetrics]
    PrefixesIterable = typing.Iterable[str]


class HttpAnalyzer:
    def __init__(
            self, access_log: AccessLog,
            prefixes: PrefixesIterable = None,
            metrics: MetricsIterable = None,
    ) -> None:
        self.access_log = access_log
        self.prefixes = list(prefixes or ['/'])
        self._metrics = list(metrics or [])

    @property
    def metrics(self) -> MetricsList:
        return self._metrics

    def add_metric(self, metric: StatisticsMetrics) -> None:
        self._metrics.append(metric)

    def analyze(self, start_date: DateOrDatetime, end_date: DateOrDatetime) -> None:
        if not self._metrics:
            return

        start_ts = timestamp(start_date, ms=True)
        end_ts = timestamp(end_date, ms=True)

        for metric in self._metrics:
            metric.initialize()

        for entry in self.access_log.get_records(start_ts, end_ts, prefixes=self.prefixes):
            for metric in self._metrics:
                metric.process_entry(entry)



from __future__ import annotations
import datetime
import typing


class Boundaries(typing.NamedTuple):
    min_ts: float
    max_ts: float


class AccessEntry(typing.NamedTuple):
    ts: int
    url: str
    method: str
    status_code: int
    request_id: str
    response_time: float


class RequestEntry(typing.NamedTuple):
    ts: int
    json: str
    request_id: str
    status_code: int


DateOrDatetime = typing.Union[datetime.date, datetime.datetime]

IndicesBoundaries = typing.Dict[str, Boundaries]



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from vuakhter.utils.types import AccessEntry


class AccessLog:
    def get_records(self, start_ts: int, end_ts: int, **kwargs: typing.Any) -> typing.Iterator[AccessEntry]:
        raise NotImplementedError()



from __future__ import annotations
import typing

from vuakhter.base.access_log import AccessLog
from vuakhter.kibana.elastic_log import ElasticLog

from vuakhter.utils.kibana import gen_access_entries

if typing.TYPE_CHECKING:
    from vuakhter.utils.types import AccessEntry


class ElasticAccessLog(ElasticLog, AccessLog):
    def __init__(self, index_pattern: str = 'filebeat-*', *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(index_pattern, *args, **kwargs)

    def gen_entries(self, index: str, **kwargs: typing.Any) -> typing.Iterator[AccessEntry]:
        prefixes = kwargs.pop('prefixes', None)
        if prefixes:
            yield from gen_access_entries(self.client, index, prefixes)



import datetime

import pytest

from vuakhter.utils.helpers import timestamp, is_valid, get_endpoint, deep_get, chunks
from vuakhter.utils.types import RequestEntry


@pytest.mark.parametrize(
    'date, ms, expected_result',
    [
        [datetime.date(2020, 4, 1), False, 1585699200],
        [datetime.date(2020, 4, 1), True, 1585699200000],
        [datetime.datetime(2020, 4, 1, 12, 12, 12), False, 1585743132],
        [datetime.datetime(2020, 4, 1, 12, 12, 12), True, 1585743132000],
    ],
)
def test_timestamp(date, ms, expected_result):
    assert timestamp(date, ms=ms) == expected_result


@pytest.mark.parametrize(
    'entry',
    [
        RequestEntry(ts=0, request_id='', status_code=200, json='{"data": {}}'),
        RequestEntry(ts=0, request_id='', status_code=400, json='{"message": "Validation error"}'),
    ],
)
def test_is_valid_passes(entry):
    assert is_valid(entry) is True


@pytest.mark.parametrize(
    'entry',
    [
        RequestEntry(ts=0, request_id='', status_code=200, json='{"data": {}, "errors": []}'),
        RequestEntry(ts=0, request_id='', status_code=400, json='{"data": {}, "errors": ["Validation error"]}'),
    ],
)
def test_is_valid_fails(entry):
    assert is_valid(entry) is False


@pytest.mark.parametrize(
    'url, expected_result',
    [
        ['/int/1234/', '/int/_INT_/'],
        ['/int/1234', '/int/_INT_'],
        ['/int/1234/?query', '/int/_INT_/'],
        ['/int/1234?query', '/int/_INT_'],
        ['/float/1234.5?query', '/float/_FLOAT_'],
        ['/uuid/8a44c1ff-0cfa-401e-b2b3-4821a677d103/?query', '/uuid/_UUID_/'],
        ['/1234/54321.0/8a44c1ff-0cfa-401e-b2b3-4821a677d103/?query', '/_INT_/_FLOAT_/_UUID_/'],
    ],
)
def test_get_endpoint(url, expected_result):
    assert get_endpoint(url) == expected_result


@pytest.mark.parametrize(
    'keys, default, expected_result',
    [
        ['level1', None, 1],
        ['level2.level2_1', None, 0],
        ['level1.level1_1', 1, 1],
        ['level2.level2_2', None, 'abcd'],
        ['level3.level3_1', 'abcd', 'abcd'],
    ],
)
def test_deep_get_returns_value(keys, default, expected_result):
    dct = {
        'level1': 1,
        'level2': {
            'level2_1': 0,
            'level2_2': 'abcd',
        },
        'level3': None,
    }

    assert deep_get(dct, keys, default=default) == expected_result


@pytest.mark.parametrize(
    'keys, default',
    [
        ['level1.level1_1', None],
        ['level3', 0],
    ],
)
def test_deep_get_returns_none(keys, default):
    dct = {
        'level1': 1,
        'level2': {
            'level2_1': 0,
            'level2_2': 'abcd',
        },
        'level3': None,
    }

    assert deep_get(dct, keys, default=default) is None


@pytest.mark.parametrize(
    'length, size, chunks_list_size, last_chunk_size',
    [
        [1, 10, 1, 1],
        [11, 10, 2, 1],
        [99, 49, 3, 1],
    ],
)
def test_chunks(length, size, chunks_list_size, last_chunk_size):
    chunk_list = list(chunks(range(1, length + 1), size=size))

    assert len(chunk_list) == chunks_list_size
    assert len(chunk_list[-1]) == last_chunk_size


def test_chunks_zero_length():
    chunk_list = list(chunks(iter([])))

    assert len(chunk_list) == 0



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING
import operator
import logging

from elasticsearch import NotFoundError
from elasticsearch_dsl import Search
from elasticsearch_dsl.query import Query, Q

from vuakhter.utils.helpers import chunks, deep_get, timestamp
from vuakhter.utils.types import Boundaries, AccessEntry, RequestEntry

if TYPE_CHECKING:
    from typing import Iterator, List, Sequence
    from elasticsearch import Elasticsearch
    from vuakhter.utils.types import IndicesBoundaries


logger = logging.getLogger(__name__)


class MatchBoolPrefix(Query):
    name = 'match_bool_prefix'


def get_timestamp(ts_str: str) -> int:
    return timestamp(
        datetime.datetime.fromisoformat(ts_str.rstrip('Z')),
        ms=True,
    )


def gen_access_entries(
        client: Elasticsearch, index: str, prefixes: Sequence[str],
) -> Iterator[AccessEntry]:
    prefix, *tail = prefixes
    query = Q('match_bool_prefix', url__original=prefix)
    for prefix in tail:
        query = query | Q('match_bool_prefix', url__original=prefix)
    search = Search(using=client, index=index).filter(query)

    logger.info('Scan %s with query %s, expect %d results', index, search.to_dict(), search.count())
    for hit in search.scan():
        hit_dict = hit.to_dict()
        try:
            yield AccessEntry(
                ts=get_timestamp(deep_get(hit_dict, '@timestamp')),
                url=deep_get(hit_dict, 'url.original'),
                method=deep_get(hit_dict, 'http.request.method'),
                status_code=int(deep_get(hit_dict, 'http.response.status_code')),
                response_time=float(deep_get(hit_dict, 'http.response.duration')),
                request_id=deep_get(hit_dict, 'nginx.access.request_id'),
            )
        except (KeyError, ValueError, TypeError):
            fail_dict = {
                'ts': deep_get(hit_dict, '@timestamp'),
                'url': deep_get(hit_dict, 'url.original'),
                'method': deep_get(hit_dict, 'http.request.method'),
                'status_code': deep_get(hit_dict, 'http.response.status_code'),
                'response_time': deep_get(hit_dict, 'http.response.duration'),
                'request_id': deep_get(hit_dict, 'nginx.access.request_id'),
            }
            logger.debug('Failed with %s', fail_dict)
            pass


def gen_request_entries(
        client: Elasticsearch, index: str, request_ids: Sequence[str],
) -> Iterator[RequestEntry]:
    search = Search(using=client, index=index)
    search = (
        search.filter('term', response__type='json_response_log')
              .filter('terms', response__request_id=request_ids)
    )

    logger.info('Scan %s with query %s, expect %d results', index, search.to_dict(), search.count())
    for hit in search.scan():
        hit_dict = hit.to_dict()
        try:
            yield RequestEntry(
                ts=get_timestamp(deep_get(hit_dict, '@timestamp')),
                json=deep_get(hit_dict, 'response.json'),
                request_id=deep_get(hit_dict, 'response.request_id'),
                status_code=int(deep_get(hit_dict, 'response.status')),
            )
        except (KeyError, ValueError, TypeError):
            fail_dict = {
                'ts': deep_get(hit_dict, '@timestamp'),
                'json': deep_get(hit_dict, 'response.json'),
                'request_id': deep_get(hit_dict, 'response.request_id'),
                'status_code': deep_get(hit_dict, 'response.status'),
            }
            logger.debug('Failed with %s', fail_dict)
            pass


def get_indices_for_timeslot(indices: IndicesBoundaries, start_ts: int, end_ts: int) -> List[str]:
    indices_names = []
    for name, boundaries in indices.items():
        if boundaries.max_ts <= start_ts or boundaries.min_ts >= end_ts:
            continue
        indices_names.append(name)
    return indices_names


def scan_indices(client: Elasticsearch, index_pattern: str) -> IndicesBoundaries:
    try:
        indices_dict = client.indices.get(index_pattern)
    except NotFoundError:
        indices_dict = {}

    indices_list = []

    for chunk in chunks(indices_dict.keys()):
        search = Search(using=client, index=','.join(chunk))[0:0]
        (
            search.aggs
                  .bucket('index', 'terms', field='_index')
                  .metric('min_ts', 'min', field='@timestamp')
                  .metric('max_ts', 'max', field='@timestamp')
        )
        result = search.execute()
        buckets = result.aggregations.index.buckets
        for bucket in buckets:
            indices_list.append(
                (bucket.key, bucket.min_ts.value, bucket.max_ts.value),
            )
    return {
        index: Boundaries(min_ts=min_ts, max_ts=max_ts)
        for index, min_ts, max_ts
        in sorted(indices_list, key=operator.itemgetter(1), reverse=True)
    }



from __future__ import annotations
import re


INT_RE = re.compile(r'^\d+\Z')
FLOAT_RE = re.compile(r'^\d+\.\d+\Z')
UUID_RE = re.compile(r'^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}\Z')

DEFAULT_CHUNK_SIZE = 10

LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s %(levelname)s %(module)s %(process)d %(thread)d %(message)s',
        },
        'simple': {
            'format': '%(levelname)s %(message)s',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'stream': 'ext://sys.stdout',
        },
    },
    'loggers': {
        'elasticsearch': {'level': 'CRITICAL', 'propagate': False},
        'requests': {'level': 'CRITICAL', 'propagate': False},
        'urllib3.connectionpool': {'level': 'CRITICAL', 'propagate': False},
    },
    'root': {
        'handlers': ['console'],
        'level': 'ERROR',
    },
}



from __future__ import annotations

import datetime
import functools
import typing
import json

from vuakhter.utils.constants import DEFAULT_CHUNK_SIZE, INT_RE, FLOAT_RE, UUID_RE, LOGGING_CONFIG

if typing.TYPE_CHECKING:
    from vuakhter.utils.types import DateOrDatetime, RequestEntry


def chunks(
    iterable: typing.Iterable[typing.Any],
    size: int = DEFAULT_CHUNK_SIZE,
) -> typing.Iterator[typing.List[typing.Any]]:
    chunk = []
    for el in iterable:
        chunk.append(el)
        if len(chunk) == size:
            yield chunk
            chunk = []
    if chunk:
        yield chunk


def deep_get(dictionary: typing.Dict, keys: str, default: typing.Any = None) -> typing.Any:
    return functools.reduce(
        lambda dct, key: dct.get(key, default) if isinstance(dct, dict) else default,
        keys.split('.'),
        dictionary,
    )


def get_endpoint(url: str) -> str:
    url = url.split('?')[0]
    parts = url.split('/')

    for idx, part in enumerate(parts):
        substituted = INT_RE.sub('_INT_', part)
        substituted = FLOAT_RE.sub('_FLOAT_', substituted)
        substituted = UUID_RE.sub('_UUID_', substituted)

        if part != substituted:
            parts[idx] = substituted

    return '/'.join(parts)


def timestamp(
        dt: DateOrDatetime, epoch: DateOrDatetime = None, ms: bool = False,
) -> int:
    epoch = epoch or type(dt)(1970, 1, 1)
    ms_multiplier = 1000 if ms else 1
    dt = dt or datetime.datetime.utcnow()
    return int((dt - epoch).total_seconds() * ms_multiplier)


def setup_logging(level: str = 'ERROR', formatter: str = 'simple') -> None:
    from logging.config import dictConfig

    if formatter in LOGGING_CONFIG['formatters'].keys():  # type: ignore
        LOGGING_CONFIG['handlers']['console']['formatter'] = formatter  # type: ignore
    LOGGING_CONFIG['root']['level'] = level  # type: ignore
    dictConfig(LOGGING_CONFIG)


def is_valid(entry: RequestEntry) -> bool:
    try:
        body = json.loads(entry.json)
    except ValueError:
        return False
    if entry.status_code == 200:
        if not (set(body.keys()) - {'data', 'meta', 'request'}):
            return True
    else:
        if not (set(body.keys()) - {'message', 'errors'}):
            return True
    return False



from __future__ import annotations
import typing

from vuakhter.utils.helpers import chunks
from vuakhter.utils.kibana import get_indices_for_timeslot, scan_indices

if typing.TYPE_CHECKING:
    from elasticsearch import Elasticsearch
    AnyIterator = typing.Iterator[typing.Any]


class ElasticLog:
    def __init__(
            self, index_pattern: str, client: Elasticsearch = None,
            *args: typing.Any, **kwargs: typing.Any,
    ):
        self.client = client or Elasticsearch(*args, **kwargs)
        self.indices = scan_indices(self.client, index_pattern)

    def gen_entries(self, index: str, **kwargs: typing.Any) -> AnyIterator:
        raise NotImplementedError()

    def get_records(self, start_ts: int, end_ts: int, **kwargs: typing.Any) -> AnyIterator:
        indices = get_indices_for_timeslot(self.indices, start_ts, end_ts)

        for chunk in chunks(indices):
            yield from self.gen_entries(','.join(chunk), **kwargs)



from __future__ import annotations
import typing
import collections

if typing.TYPE_CHECKING:
    from vuakhter.utils.types import AccessEntry
    StatisticsMapping = typing.Dict[typing.Any, typing.Any]


class StatisticsMetrics:
    _description = 'Some metrics'

    def __init__(self) -> None:
        self.initialize()

    @property
    def description(self) -> str:
        return self._description

    @property
    def statistics(self) -> StatisticsMapping:
        return self._statistics

    def initialize(self) -> None:
        self._statistics: StatisticsMapping = collections.defaultdict(int)

    def finalize(self) -> None:
        pass

    def process_entry(self, entry: AccessEntry) -> None:
        raise NotImplementedError()

    def report(self) -> str:
        report = [self.description, '-' * len(self.description), f'{self._statistics}']
        for key, value in self._statistics.items():
            report.append(f'{key} {value}')
        return '\n'.join(report)



from __future__ import annotations
import typing
import math

from vuakhter.metrics.base import StatisticsMetrics
from vuakhter.utils.helpers import get_endpoint, is_valid

if typing.TYPE_CHECKING:
    from vuakhter.base.requests_log import RequestsLog
    from vuakhter.utils.types import AccessEntry


class MethodCounter(StatisticsMetrics):
    description = 'Response counts by method'

    def process_entry(self, entry: AccessEntry) -> None:
        self._statistics[entry.method] += 1


class EndpointCounter(StatisticsMetrics):
    description = 'Response counts by endpoint'

    def process_entry(self, entry: AccessEntry) -> None:
        endpoint = get_endpoint(entry.url)
        self._statistics[endpoint] += 1


class StatusCounter(StatisticsMetrics):
    description = 'Response counts by status'

    def process_entry(self, entry: AccessEntry) -> None:
        self._statistics[entry.status_code] += 1


class ComplexCounter(StatisticsMetrics):
    description = 'Response counts by endpoint, method, status'

    def process_entry(self, entry: AccessEntry) -> None:
        key = (get_endpoint(entry.url), entry.method, entry.status_code)
        self._statistics[key] += 1


class ResponseTimeCounter(StatisticsMetrics):
    description_prefix = 'Response counts by time took'

    def __init__(self, fraction: float = 0.50):
        super().__init__()
        self.fraction = fraction

    @property
    def description(self) -> typing.Any:
        return f'{self.description_prefix} fraction by {self.fraction} ms'

    def process_entry(self, entry: AccessEntry) -> None:
        response_time = math.ceil(entry.response_time / self.fraction) * self.fraction
        self._statistics[response_time] += 1


class SchemaValidatorCounter(StatisticsMetrics):
    description = 'Number of valid with schema responses'

    def __init__(self, requests_log: RequestsLog):
        super().__init__()
        self.requests_log = requests_log
        self.min_ts = 0
        self.max_ts = 0
        self.check_ids: typing.List[str] = []

    def process_buffer(self) -> None:
        if not self.check_ids:
            return
        valid_cnt, invalid_cnt = 0, 0
        for record in self.requests_log.get_records(self.min_ts, self.max_ts, request_ids=self.check_ids):
            if is_valid(record):
                valid_cnt += 1
            else:
                invalid_cnt += 1

        self._statistics['valid'] += valid_cnt
        self._statistics['invalid'] += invalid_cnt
        self._statistics['missed'] += len(self.check_ids) - (valid_cnt + invalid_cnt)

        self.min_ts = 0
        self.max_ts = 0
        self.check_ids = []

    def process_entry(self, entry: AccessEntry) -> None:
        if not self.min_ts or entry.ts < self.min_ts:
            self.min_ts = entry.ts
        if not self.max_ts or entry.ts > self.max_ts:
            self.max_ts = entry.ts
        self.check_ids.append(entry.request_id)
        if len(self.check_ids) >= 100:
            self.process_buffer()

    def finalize(self) -> None:
        self.process_buffer()

    def report(self) -> str:
        valid, invalid, missed = self._statistics['valid'], self._statistics['invalid'], self._statistics['missed']
        validated = (valid + invalid)
        rate = 0.0
        if validated:
            rate = 100.0 * self._statistics['valid'] / validated
        return f'Valid requests {rate:.2f}% ({valid} out of {validated}, {missed} missed)'
